from ._version import __version__

from . import evolution
from .evolution import Evolution
# from .creatures import EvogressionCreature
from . import groups
from .groups import evolution_group, parameter_usage, output_usage, random_population, Population
